/*
 *  VCFto012s.cpp
 *  vcfReader
 *
 *  Created by Youna Hu on 2/3/11.
 *  Copyright 2011 U of Michigan. All rights reserved.
 *
 */

#include "VCFto012s.h"

void VcfTransform::Init(String &vcfName, String &subjectsFile){
	m_vcfName = vcfName;
	m_subjectFile = subjectsFile;
	fileStep = 0;
	searchNonsyn = ";NON_SYN";
	searchStop = ";STOP";
	genoSum = 0;
	nSample = 0;
	genoStartCol = 9; // genotyep data starts at the 10th column
	idCol = 2; // ID is at the 3rd column
	refCol = 3; // reference allele is the 4th column
	altCol = 4; // alternative allele is the 5th column
	annotationCol = 7; // annotation is the 8th column 
	sampleByPos = false; // sampleByPos: rows -- samples, cols -- Positions
	numPos = 0;
	UpdateSubjects();
	rowKeep.clear();
	
}

VcfTransform::VcfTransform(String &vcfName, String &subjectsFile){
	Init(vcfName, subjectsFile);
}


VcfTransform::~VcfTransform(){
	;
}

bool VcfTransform::IsMissing(){
	if (alleleToken[0] != "0" && alleleToken[0] != "1" && alleleToken[1] != "0" && alleleToken[1] != "1") {
		return (true);
	} else {
		return (false);
	}
}

// update subjectColumn, subjectSelected
void VcfTransform::UpdateSubjects(){
	// read in all subjects
	FILE *vcfIn = fopen(m_vcfName, "r");
	if (vcfIn == NULL) {
		error("Error in opening file (%s)\n",m_vcfName.c_str());
	}
	fileStep = 0;
	while (!feof(vcfIn)) {
		fileEntry.ReadLine(vcfIn);
		if (fileEntry.IsEmpty()) {
			error("Error: File %s has an empty line (%d)\n",m_vcfName.c_str(),fileStep);
		}
		// find the line starts with $CHROM and read the subject IDs
		if (fileEntry.Left(6) == "#CHROM") {
			//	printf("Found %s\n",fileEntry.Left(6).c_str());
			fileToken.AddTokens(fileEntry, "\t");
			for (int i = genoStartCol; i < fileToken.Length(); i++) {
				subjectColumn.insert(std::map<String,int>::value_type(fileToken[i],i));
				subjectSelected.push_back(fileToken[i].c_str());
			}
			break;
		}
	}
	fclose(vcfIn);
	// The previous code select everybody, if there is a valid subjectFile, udpate this
	if (m_subjectFile != "") {
		subjectSelected.clear();
		DataFromFiles subjectIDs;
		subjectIDs.IDReadAsString(m_subjectFile);
		subjectSelected = subjectIDs.GetIndvsIDs();
	}
	nSubjects = subjectSelected.size();
	printf("Note: %d subjects are included.\n",nSubjects);
}

bool VcfTransform::AnnotationSearch(String & fileEntryIn, int idx_to_search, String & searchKeys){
	// check how many keys to search
	int numItemSearch;
	// borrow the memory of genoToken to use for search token
	genoToken.Clear();
	genoToken.AddTokens(searchKeys,":");
	numItemSearch = genoToken.Length();
	// search the to-search string
	fileToken.Clear();
	fileToken.AddTokens(fileEntryIn, "\t");
	annotationInfo.clear();
	annotationInfo = fileToken[idx_to_search].c_str();
	printf("annotationInfo = %s\n",annotationInfo.c_str());
	bool findTrue = true;
	for (int i = 0; i < numItemSearch; i ++) {
		findTrue = findTrue && (annotationInfo.find(genoToken[i].c_str()) != std::string::npos);
		printf("number %d - findTrue = %d\n",i, findTrue);
	}
	// clear the tokens
	genoToken.Clear();
	fileToken.Clear();
	return (findTrue);
}


// search the chromosome number and if the position fall in between the start, end range
bool VcfTransform::PositionRangeSearch(String &fileEnetryIn, String chrNumber, int rangeStart, int rangeEnd){	
	fileToken.Clear();
	fileToken.AddTokens(fileEnetryIn, "\t");
	
	int cmpTrue = true;
	
	// check the position number
	if (fileToken[1].AsInteger() < rangeStart || fileToken[1].AsInteger() > rangeEnd) {
		return (false);
	}	
	
	// check the chromosome number
	if(fileToken[0].Length() >= 4){ // if the length is greater than 3, that means it uses chr+num
		cmpTrue = cmpTrue && (fileToken[0].SubStr(3) == chrNumber);
	} else {
		cmpTrue = cmpTrue && (fileToken[0] == chrNumber);
	} // need to be careful about X, Y chromosomes
	
	return (cmpTrue);
}

void VcfTransform::VCFvariantsInclude(int posAll){	
	if (posAll == 1) { // if every position is included, then just go over the positions
		FILE *vcfIn = fopen(m_vcfName, "r");
		if (vcfIn == NULL) {
			error("Error in opening file (%s)\n",m_vcfName.c_str());
		}
		numPos = 0;
		fileStep = 0;
		while (!feof(vcfIn)){
			fileEntry.ReadLine(vcfIn);
			if (fileEntry.IsEmpty()) {
				error("Error: File %s has an empty line (%d)\n",m_vcfName.c_str(),fileStep);
			}
			fileStep ++;
			if (fileEntry.Left(1) != "#"){
				numPos ++; // count the number of lines without # (the position lines)
				rowKeep.insert(fileStep);
			}
		}
		fileEntry.Clear();
		fclose(vcfIn);
	}
	else if (posAll == 0){ // if only nonsynonymous position is included
		FILE *vcfIn = fopen(m_vcfName, "r");
		if (vcfIn == NULL) {
			error("Error in opening file (%s)\n",m_vcfName.c_str());
		}
		numPos = 0;
		fileStep = 0;
		while (!feof(vcfIn)){
			fileEntry.ReadLine(vcfIn);
			if (fileEntry.IsEmpty()) {
				error("Error: File %s has an empty line (%d)\n",m_vcfName.c_str(),fileStep);
			}	
			fileStep ++;
			fileToken.Clear();
			if (fileEntry.Left(1) != "#"){ 
				fileToken.AddTokens(fileEntry, "\t");
				annotationInfo = fileToken[annotationCol].c_str();
				if (annotationInfo.find(searchNonsyn) != std::string::npos 
								|| annotationInfo.find(searchStop) != std::string::npos){ // check the nonsyn stop position
							numPos ++; // count the number of lines without # (the position lines)
							rowKeep.insert(fileStep); // update the map to record which row
						}
			}
			fileEntry.Clear();
			fileToken.Clear();
		}
		fclose(vcfIn);
	} else {
		error("Invalid variantChoice. Pick 1 or 0. \n");
	}	
}

void	VcfTransform::VCFvariantsInclude(int posAll, String &vGroups){
	genoToken.Clear();
	genoToken.AddTokens(vGroups, ":");
	if (genoToken.Length() != 3) {
		error("Invalid variant group specification! Use two numbers seperated by a colon, for example, 1:100:200.\n");
	}
	String chrNum = genoToken[0];
	
	int vStart = genoToken[1];
	int vEnd = genoToken[2];
	if (vStart > vEnd) {
		error("The start position is greater thant the end position. Check!\n");
	}
	
	printf("Susetting criterion: chromosome number: %s, start position: %d, end position: %d\n",
				 chrNum.c_str(),vStart,vEnd);
	
	if (posAll == 1) { // if every position is included, then just go over the positions
		FILE *vcfIn = fopen(m_vcfName, "r");
		if (vcfIn == NULL) {
			error("Error in opening file (%s)\n",m_vcfName.c_str());
		}
		numPos = 0;
		fileStep = 0;
		while (!feof(vcfIn)){
			fileEntry.ReadLine(vcfIn);
			if (fileEntry.IsEmpty()) {
				error("Error: File %s has an empty line (%d)\n",m_vcfName.c_str(),fileStep);
			}
			fileStep ++;
			if (fileEntry.Left(1) != "#" && PositionRangeSearch(fileEntry, chrNum, vStart, vEnd)) { //check position range
				numPos ++; // count the number of lines without # (the position lines)
				rowKeep.insert(fileStep);
			}
		}
		fileEntry.Clear();
		fclose(vcfIn);
	}
	else if (posAll == 0){ // if only nonsynonymous position is included
		FILE *vcfIn = fopen(m_vcfName, "r");
		if (vcfIn == NULL) {
			error("Error in opening file (%s)\n",m_vcfName.c_str());
		}
		numPos = 0;
		fileStep = 0;
		while (!feof(vcfIn)){
			fileEntry.ReadLine(vcfIn);
			if (fileEntry.IsEmpty()) {
				error("Error: File %s has an empty line (%d)\n",m_vcfName.c_str(),fileStep);
			}	
			fileStep ++;
			fileToken.Clear();
			if (fileEntry.Left(1) != "#"){ 
				fileToken.AddTokens(fileEntry, "\t");
				annotationInfo = fileToken[annotationCol].c_str();
				if (PositionRangeSearch(fileEntry, chrNum, vStart, vEnd) //check position range
						&& (annotationInfo.find(searchNonsyn) != std::string::npos 
								|| annotationInfo.find(searchStop) != std::string::npos)){ // check the nonsyn stop position
							numPos ++; // count the number of lines without # (the position lines)
							rowKeep.insert(fileStep); // update the map to record which row
						}
			}
			fileEntry.Clear();
			fileToken.Clear();
		}
		fclose(vcfIn);
	} else {
		error("Invalid variantChoice. Pick 1 or 0. \n");
	}	
}

void VcfTransform::UpdateColumnSelected(String & subjectsFile, String &genofileName){
	
	FILE *genoIndv = fopen(genofileName + ".012.indv", "w");	
	// columnsSelected would be either the size of subjectSelected or the total number of subject in the VCF file
	if (subjectsFile != "") { // if not all subjects are included
		printf("Note: All individuals are included in the tranformation.\n");
		for (int i = 0; i < subjectSelected.size(); i++) {
			std::map<String,int>::iterator idx = subjectColumn.find(subjectSelected[i]);
			columnsSelected.push_back(idx->second);
			fprintf(genoIndv, "%s\n",(idx->first).c_str()); // it shouldn't be subjectSelected[i] b/c the subjectSelected might have more subject
		}
	} else {
		printf("Note: File (%s) specifies the subjects inlcuded for the transformation.",subjectsFile.c_str());
		for	(std::map<String,int>::iterator i = subjectColumn.begin(); i != subjectColumn.end(); i++){
			columnsSelected.push_back(i->second); // note: since we've done this, need to ouptut the order of subject as in subjectColumn, and it should be the same order as the original vcf subject row header
			fprintf(genoIndv, "%s\n",(i->first).c_str());
		}
	}
}


void VcfTransform::VCF2TransformedGeno(String &genofileName, String &subjectsFile, int posAll, String &geneDef){
	// Go though all variants and determine whether to include each one and how many rows are included
	printf("Check the number of variants to be included...\n");
	if (geneDef != "") {
		VCFvariantsInclude(posAll, geneDef);
	} else {
		VCFvariantsInclude(posAll);
	}
	printf("Note: %d variants are included.\n",numPos);
	printf("Done.\n");
	
	// update columns to be kept
	UpdateColumnSelected(subjectsFile,genofileName);
	
	// allocate a matrix of size numSubject by numPos
	Matrix tmpMatGenoStore(nSubjects,numPos);
	tmpMatGenoStore.Set(-1);
	
	// read in the genotype matrix
	FILE *vcfIn = fopen(m_vcfName, "r");
	FILE *genoOut = fopen(genofileName + ".012","w");
	FILE *genoPos = fopen(genofileName + ".012.pos", "w");
	FILE *genoAF = fopen(genofileName + ".012.frq", "w");
	
	// output allele frequency file with header: CHROM POS REF:Frequency ALT:Frequency
	fprintf(genoAF, "%s\t%s\t%s\t%s\t%s\n", "chrm", "pos","rsName", "ref","alt");
	
	if (vcfIn == NULL) {
		printf("Error in opening file (%s)\n",m_vcfName.c_str());
		throw;
	}
	
	if (genoOut == NULL){
		printf("Error in opening file (%s)\n",genofileName.c_str());
		throw;
	}
	
	fileStep = 0; // this fileStep is used to step into the tmp geno matrix
	int matStep = 0;
	
	while (!feof(vcfIn)) {
		fileEntry.ReadLine(vcfIn);
		if (fileEntry.IsEmpty()) {
			printf("Error: File %s has an empty line (%d)\n",m_vcfName.c_str(),fileStep);
			throw;
		}
		fileStep ++;
		fileToken.Clear();
		genoToken.Clear();
		alleleToken.Clear();
		
		if (rowKeep.find(fileStep) != rowKeep.end() && fileEntry.Left(1) != "#") {
			fileToken.AddTokens(fileEntry, "\t");
			// write the position file
			fprintf(genoPos, "%s\t%s\n", fileToken[0].c_str(), fileToken[1].c_str());
			// write the genotypes and initialize the genoSum counts
			genoSum = 0;
			nSample = 0;
			for (int i = 0; i < columnsSelected.size(); i++) {
				// get the genotype
				genoToken.AddTokens(fileToken[columnsSelected[i]], ":");
				alleleToken.AddTokens(genoToken[0], "/"); 
				
				// record the genotype -- check if it is misisng first
				if (!IsMissing()) {
					genoTypeNum = alleleToken[0].AsInteger() + alleleToken[1].AsInteger();	
					genoSum += genoTypeNum;
					nSample += 1;
				} else {
					genoTypeNum = -1;
				}
				
				// write out the genotype into the matrix
				tmpMatGenoStore[i][matStep] = genoTypeNum;
				genoToken.Clear();
				alleleToken.Clear();
			}	
			matStep ++;
			// write out the allele frequency (ID, ref, alt, af)
			double dSample = static_cast<double>(nSample);
			double dGenoSum = static_cast<double>(genoSum);
			double af = 0.5*dGenoSum/dSample;
			fprintf(genoAF, "%s\t%s\t%s\t%s:%1.5e\t%s:%1.5e\n", fileToken[0].c_str(), fileToken[1].c_str(), fileToken[2].c_str(),		
							fileToken[refCol].c_str(), 1-af, fileToken[altCol].c_str(), af);		
		} 
		
		fileEntry.Clear();
		fileToken.Clear();
	}			
	fclose(genoPos);
	fclose(vcfIn);
	fclose(genoAF);	
	
	// now, output the 012 matrix into file -- we could skip this step and use the matrix in the RVtest
	// but for now, we decide to store the matrix first and re-read it.
	for (int i = 0; i < nSubjects; i ++) {
		for (int j = 0; j < numPos; j ++) {
			fprintf(genoOut, "%1.0f\t", tmpMatGenoStore[i][j]);
		}
		fprintf(genoOut, "\n");
	}
	fclose(genoOut);	
}	





