#include "StringBasics.h"
#include "StringArray.h"
#include "Parameters.h"
#include "FunctionTimer.h"
#include <iostream>
#include "annotateLog.h"
#include "VCFto012s.h"

int main (int argc, char ** const argv) {
	String inputVcfName("LDL.test.vcf");
	String inputLog(""); //LDL.test.log
	String output012("subsetGeno");
	String geneList(""); // LDL.geneList.txt
	String subsetID(""); // LDL.AA.txt
	// choose variants to output -
	// default variant choice is NonsynStop
	// available variantsChoice is {0-NonsynStop,1-All}
	int variantsChoice = 0;
	// available variantsGroup by chr#:start:end, eg, "1:1000:2000"
	String variantsGroup(""); 
	ParameterList pl;
	
	//command line specification
	// todo: 1, convert ped file into 012 matrix
	BEGIN_LONG_PARAMETERS(options)
		LONG_PARAMETER_GROUP("Input files")
		LONG_STRINGPARAMETER("vcf", &inputVcfName)
		LONG_STRINGPARAMETER("log",&inputLog)
		LONG_STRINGPARAMETER("IDfile", &subsetID)
		LONG_PARAMETER_GROUP("Subsetting choices")
		LONG_INTPARAMETER("All", &variantsChoice)	//	All:1, NonsynStop:0
		LONG_STRINGPARAMETER("positionRange", &variantsGroup)
		LONG_PARAMETER_GROUP("Output files")
		LONG_STRINGPARAMETER("outputPrefix",&output012)
		LONG_STRINGPARAMETER("outputGeneList",&geneList)
	END_LONG_PARAMETERS();
	
	pl.Add((new LongParameters("Data File PREPARATION",options))->SetPrecision(3));
	pl.Read(argc, argv);
	pl.Status();
	
	if (inputLog != "") {
		GeneMap testLog;
		printf("Read in the annotation information from file (%s)......\n",inputLog.c_str());
		printf("Write out the gene list into file (%s)......\n",geneList.c_str());
		AnnotateLog::Output_GeneInfo(inputLog,testLog,geneList);
		printf("Done.\n");		
	}
	
	// variantCallForm initalization
	VcfTransform test(inputVcfName,subsetID);
	FunctionTimer recordTime;
	printf("Read in the vcf file (%s) ......\n", inputVcfName.c_str());
	printf("Write out the 012 genotype matrix files %s (.012, .012.pos, .012.indv, .012.frq) \n",output012.c_str());
	test.VCF2TransformedGeno(output012, subsetID, variantsChoice, variantsGroup);
	printf("Done.\n");

	std::cout << "It takes " << recordTime.StopTime()/1e6 << " seconds" << std::endl;
	return 0;
}
